/* 
 * Evorobot - A software for running evolutionary robotics experiments
 * Copyright (C) 2000 Stefano Nolfi. - Institute of Psychology - CNR - Rome 
 * nolfi@ip.rm.cnr.it - http://gral.ip.rm.cnr.it/nolfi 

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 * file - learning.c
 */

#include "defs.h" 


/*
 * Plain Hebb learning rule (called by 00)  
 */

float hebb(float x, float y, float w)

{
  return (x*y*(1-w));
}

/*
 * Presynaptic learning rule (called by 10)
 */
float presynaptic(float x, float y, float w)

{
  return (y*(1-w)*x+(y-1)*w*x);
}


/*
 * Postsynaptic learning rule (called by 01)   
 */
float postsynaptic(float x, float y, float w)

{
  return (x*(1-w)*y+(x-1)*w*y);
}


/*
 * Continuous-covariance learning rule (called by 11) 
 */
float ccov(float x, float y, float w)

{
  double temp;
  extern double tanh();
  extern double fabs();

  temp=tanh( (2.0*(2.0*(1.0-fabs((double)(x-y)))-1)) );
  if(temp>0.0)
    return (float)(temp*(1-w));
    else
      return (float)(temp*w);
}



/*
 *  update weights by using hebbian learning
 */  
void
hebbian_update(struct individual *pind)

{

	int   *ldp_b;
	int   *ldp_t;
	float **layerp_b;
	float **layerp_t;
	float **weight_p;
	float **bias_p;
	int   **wrule;
	float **wrate;
	int   **brule;
	float **brate;
	float *lp_t;
	float *lp_b;
	float *w_p;
	float *b_p;
	int   *w_ru;
	float *w_ra;
        int   *b_ru;
	float *b_ra;
	int   i;
	int   b;
	int   t;
	int   sign;

	/*
	 * do the rest for nl-1 (==nlayers-2);  we only have
	 * (nlayers-1) connections to change, and we've already done
	 * top set
	 */
	 ldp_b    = pind->layer_descp;
	 ldp_t    = (pind->layer_descp + 1);
	 layerp_b = pind->layerp;
	 layerp_t = (pind->layerp + 1);
	 weight_p = pind->weightp;
	 wrule    = pind->wrules;
	 wrate    = pind->wrates;
	 brate    = pind->brates;
	 brule    = pind->brules;
	 bias_p   = (pind->biasp + 1);
	 for (i = 0; i < (pind->nlayers - 1); i++, ldp_t++, ldp_b++, layerp_b++, layerp_t++, weight_p++, bias_p++, wrate++, wrule++, brate++, brule++) 
	     {
	       w_p  = *weight_p;
	       b_p = *bias_p;
	       w_ru = *wrule;
	       w_ra = *wrate;
               b_ru = *brule;
	       b_ra = *brate;

	       for (t = 0, lp_t = *layerp_t; t < *ldp_t; t++, lp_t++, b_p++, b_ru++, b_ra++) 
                {
		  for(b =0, lp_b = *layerp_b; b < *ldp_b; b++, lp_b++, w_p++, w_ru++, w_ra++) 
		   {
		     if (*w_p > 0)
		       {
			sign = 0;
		       }
		       else
		       {
			sign = 1;
			*w_p = (float) 0.0 - *w_p;
		       }
                     if (*w_ru == 0)
			 *w_p += (hebb(*lp_b, *lp_t, *w_p) * *w_ra);
                     if (*w_ru == 1)
			 *w_p += (presynaptic(*lp_b, *lp_t, *w_p) * *w_ra);
                     if (*w_ru == 2)
			 *w_p += (postsynaptic(*lp_b, *lp_t, *w_p) * *w_ra);
                     if (*w_ru == 3)
			 *w_p += (ccov(*lp_b, *lp_t, *w_p) * *w_ra);
		     if (sign == 1)
			*w_p = (float) 0.0 - *w_p;
		   }
		     if (*b_p > 0)
		       {
			sign = 0;
		       }
		       else
		       {
			sign = 1;
			*b_p = (float) 0.0 - *b_p;
		       }
                  if (*b_ru == 0)
			 *b_p += (hebb((float) 1.0, *lp_t, *b_p) * *b_ra);
                  if (*b_ru == 1)
			 *b_p += (presynaptic((float) 1.0, *lp_t, *b_p) * *b_ra);
                  if (*b_ru == 2)
			 *b_p += (postsynaptic((float) 1.0, *lp_t, *b_p) * *b_ra);
                  if (*b_ru == 3)
			 *b_p += (ccov((float) 1.0, *lp_t, *b_p) * *b_ra);
		  if (sign == 1)
			*b_p = (float) 0.0 - *b_p;
	        }
	   }

}



/*
 * back_propagation
 * computes error and deltas of the output layer
 * it takes the output layer, the size of the output layer, 
 *  the teaching input pointer and the delta of the last layer, 
 *  then for recurrent deltas the number of inputs and the delta of inputs
 */
float
delta_out(float *outp, int outsize, float *tarp, float *outdeltap, int ninputs, float *inpdeltap)

{
	register int i;
	register float out;
	register float diff;
        float	global_error;


	global_error = (float) 0.0;
	for (i=0; i<outsize; i++) 
	    {
	     out = *outp;
	     if (*tarp != -999.0)
	       diff = (*tarp - out);
              else
	       diff = (float) 0.0;
	     *outdeltap = diff * out * ((float) 1.0 - out);
	     global_error += diff*diff;
	     outdeltap++;
	     outp++;
	     tarp++;
	    }
	return(global_error);
}


/*
 * computes delta for all the layer but the output.
 * for each layer of weights it takes :
 *  the size of units of the receiving layer, deltas of the receiving layer, 
 *  the layer weights, the units of the sending layer, the size of the 
 *  sending layer and delta of the receiving layer.
 */
void
delta(int t_size, float *t_deltap, float *t_weightp, float *b_actp, int b_size, float *b_deltap)

{
	register float tmp;
	register float *tdp;
	register int t;
	register int b;

	for (b=0; b<b_size; b++) 
	    {
	      tmp = (float) 0.0;
	      for (t=0, tdp=t_deltap; t<t_size; t++, tdp++) 
		  {
		    tmp += *tdp * *(t_weightp + (t*b_size) + b);
		  }
	      *b_deltap = ((float) 1.0 - *b_actp) * *b_actp * tmp;
	      b_deltap++;
	      b_actp++;
	   }
}


/* 
 * update weights 
 */

void
upd_weight(int t_size, float *t_deltap, float *t_weightp, float *t_biasp, float *b_actp, int b_size)

{
	register float *bap;
	register float *endp;
	register int t;
	register float dp_rate;

	endp = b_actp + b_size;
	for (t=0; t<t_size; t++) 
	    {
	      bap = b_actp;
	      dp_rate = *t_deltap++ * rate;
	      while (bap < endp) 
		{
		  *t_weightp++ += *bap++ * dp_rate;
		}
	      *t_biasp++ += dp_rate;
	    }
}






/*
 *  update weights
 */  
void
update_net(struct individual *pind, float *teachp)

{

	float **lp;
	float **wp;
	float **dwp;
	float **iwp;
	float **dp;
	float **dp0;
	float **bp;
	float **dbp;
	float **ibp;
	int   *ldp;
	int   *ldp0;
	register int i;
	int     nl;

	  nl = pind->nlayers - 1;
	 /*
	  * do layers in reverse; output layer is done differently
	  */
	  ldp = pind->layer_descp + nl;
	  lp  = pind->layerp + nl;
	  wp  = pind->weightp + (nl - 1);
	  dp  = pind->deltap + nl;
	  dp0 = pind->deltap;
	  ldp0 = pind->layer_descp;
	  pind->global_error = delta_out(*lp, *ldp, teachp, *dp, *ldp0, *dp0);
	  if (nolearn == 1)
	     return;
	   /*
	    * do the rest for nl-1 (==nlayers-2);  we only have
	    * (nlayers-1) connections to change, and we've already done
	    * top set
	    */
	    for (i = 0; i < (nl - 1); i++) 
	     {
	        delta(*ldp, *dp, *wp, *(lp - 1), *(ldp - 1), *(dp - 1));
		lp--;
		ldp--;
		dp--;
		wp--;
	      }
	    /*
	     * go back and update the weights and biases, given the deltas
	     * we just computed
	     */
	     ldp = pind->layer_descp + nl;
	     lp  = pind->layerp + nl;
	     wp  = pind->weightp + (nl - 1);
	     dp  = pind->deltap + nl;
	     bp  = pind->biasp + nl;
	     for (i = 0; i < nl; i++) 
	      {
	       upd_weight(*ldp, *dp, *wp, *bp, *(lp - 1), *(ldp - 1));
	       ldp--;
	       dp--;
	       wp--;
	       dwp--;
	       iwp--;
	       bp--;
	       dbp--;
	       ibp--;
	       lp--;
	      }
	

}




